"""TCP socket input provider.

It connects to a TCP server,
listens for `PROVIDE:` messages to receive prompts,
and sends `USE:` messages to pass the response.
"""

import logging
import socket
import threading
import time
from types import TracebackType
from typing import Optional, Type

END_OF_MESSAGE = b"\r\n"
LOGGER = logging.getLogger("tcp::provider")


class InputProviderThread(threading.Thread):
    """Input provider thread."""

    def __init__(
        self,
        host: str,
        port: int,
        response: Optional[str],
        timeout: Optional[float] = None,
    ) -> None:
        """Create a new input provider thread.

        Parameters
        ----------
        host : str
            The host to connect to.
        port : int
            The port to connect to.
        response : str, optional
            The response to send.
        timeout : float, optional
            The timeout for the provider, by default None (no timeout).
        """
        super().__init__(
            name="InputProviderThread",
            daemon=True,
            target=self.run,
        )
        self.host = host
        self.port = port
        self.timeout = timeout
        self._response = response
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        self.running = False

    @property
    def response(self) -> Optional[str]:
        """Get the response."""
        return self._response

    @response.setter
    def response(self, response: str) -> None:
        """Set the response.

        Parameters
        ----------
        response : str
            The response to set.
        """
        self._response = response

    def _run_no_timeout(self) -> None:
        """Start the provider."""
        self.socket.connect((self.host, self.port))
        self.socket.sendall("PROVIDER\r\n".encode("utf-8"))
        self.running = True
        while self.running:
            data = b""
            while not data.endswith(END_OF_MESSAGE):
                try:
                    data += self.socket.recv(1024)
                except BaseException:  # pylint: disable=broad-except
                    break
            if not data:
                break
            if data.startswith(b"PROVIDE:"):
                prompt = data[len(b"PROVIDE:") : -len(END_OF_MESSAGE)]
                LOGGER.debug("Got a prompt: %s", prompt)
                if self.response:
                    message = f"USE:{self.response}" + "\r\n"
                    self.socket.sendall(message.encode("utf-8"))
        self.socket.close()
        self.running = False

    def _run_with_timeout(self, timeout: float) -> None:
        """Start the provider with a timeout."""
        self.socket.connect((self.host, self.port))
        self.socket.sendall("PROVIDER\r\n".encode("utf-8"))
        self.running = True
        start_time = time.monotonic()
        while self.running:
            data = b""
            while time.monotonic() - start_time < timeout:
                try:
                    data += self.socket.recv(1024)
                except BaseException:  # pylint: disable=broad-except
                    break
                if data.endswith(END_OF_MESSAGE):
                    break
            if not data:
                break
            if data.startswith(b"PROVIDE:"):
                prompt = data[len(b"PROVIDE:") : -len(END_OF_MESSAGE)]
                LOGGER.debug("Got a prompt: %s", prompt)
                if self.response:
                    message = f"USE:{self.response}" + "\r\n"
                    self.socket.sendall(message.encode("utf-8"))
        self.socket.close()
        self.running = False

    def run(self) -> None:
        """Run the provider."""
        if self.timeout is None or self.timeout < 1:
            self._run_no_timeout()
        else:
            self._run_with_timeout(self.timeout)


class InputProviderWrapper:
    """Input provider wrapper."""

    thread: Optional[InputProviderThread] = None

    def __init__(
        self,
        host: str,
        port: int,
        response: Optional[str] = None,
        timeout: Optional[float] = None,
    ) -> None:
        """Create a new input provider wrapper.

        Parameters
        ----------
        host : str
            The host to connect to.
        port : int
            The port to connect to.
        response : str, optional
            The response to send.
        timeout : float, optional
            The timeout for the provider, by default None (no timeout).
        """
        self._host = host
        self._port = port
        self._timeout = timeout
        self._response = response

    @property
    def response(self) -> Optional[str]:
        """Get the response."""
        if self.thread:
            return self.thread.response
        return self._response

    def start(self) -> None:
        """Start the provider."""
        self.thread = InputProviderThread(
            self._host, self._port, self._response, self._timeout
        )
        self.thread.daemon = True
        self.thread.start()

    def set_response(self, response: str) -> None:
        """Set the response.

        Parameters
        ----------
        response : str
            The response to set.
        """
        self._response = response
        if self.thread:
            self.thread.response = response

    def stop(self) -> None:
        """Stop the provider."""
        if self.thread:
            self.thread.running = False
            self.thread.join(timeout=1)
            del self.thread
            self.thread = None


class TCPProvider:
    """Input provider."""

    _wrapper: Optional[InputProviderWrapper] = None

    def __init__(
        self,
        host: str,
        port: int,
        response: Optional[str] = None,
        timeout: Optional[float] = 30,
    ) -> None:
        """Create a new input provider.

        Parameters
        ----------
        host : str
            The host to connect to.
        port : int
            The port to connect to.
        response : str, optional
            The response to send.
        timeout : float, optional
            The timeout for the provider, by default 30.
        """
        self._host = host
        self._port = port
        self._response = response
        self._timeout = timeout
        self._start_called = False
        self._init_wrapper()

    def __enter__(self) -> "TCPProvider":
        """Enter the context."""
        self.start()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Exit the context."""
        self.stop()

    @property
    def thread(self) -> Optional[InputProviderThread]:
        """Get the thread."""
        if not self._wrapper:
            return None
        return self._wrapper.thread

    @property
    def response(self) -> Optional[str]:
        """Get the response."""
        if self._wrapper:
            return self._wrapper.response
        return self._response

    def _init_wrapper(self) -> None:
        """Initialize the wrapper."""
        self._wrapper = InputProviderWrapper(
            host=self._host,
            port=self._port,
            response=self._response,
            timeout=self._timeout,
        )

    def is_running(self) -> bool:
        """Check if the provider is running.

        Returns
        -------
        bool
            True if the provider is running, False otherwise.
        """
        if not self._wrapper:
            return False
        if not self.thread:
            return False
        return self.thread.running is True

    def wait(self, timeout: float) -> None:
        """Wait until the provider is running.

        Parameters
        ----------
        timeout : float
            The timeout to wait.
        """
        start_time = time.time()
        while not self.is_running():
            if time.time() - start_time > timeout:
                break
            time.sleep(1)

    def start(self) -> None:
        """Start the provider.

        Raises
        ------
        RuntimeError
            If the wrapper is not initialized.
        """
        # avoid starting the provider multiple times
        # (if the wrapped thread has not yet started)
        if self._start_called is True:
            return
        self._start_called = True
        if self.is_running():
            return
        if not self._wrapper:
            raise RuntimeError("Wrapper not initialized")
        self._wrapper.start()

    def set_response(self, response: str) -> None:
        """Set the response.

        Parameters
        ----------
        response : str
            The response to set.

        Raises
        ------
        RuntimeError
            If the wrapper is not initialized.
        """
        if not self._wrapper:
            raise RuntimeError("Wrapper not initialized")
        self._wrapper.set_response(response or "\n")

    def stop(self) -> None:
        """Stop the provider."""
        self._start_called = False
        if not self.is_running():
            return
        if self._wrapper:
            self._wrapper.stop()
        del self._wrapper
        self._init_wrapper()

    def restart(self) -> None:
        """Restart the provider."""
        self.stop()
        self.start()



"""TCP socket input consumer.

It connects to a TCP server,
listens for `INPUT:` messages to get the user's input,
and sends `REQUEST:` messages to prompt the user,
"""

import socket
import time
from types import TracebackType
from typing import Optional, Type

END_OF_MESSAGE = b"\r\n"


class TCPConsumer:
    """TCP socket input consumer."""

    def __init__(
        self, host: str, port: int, timeout: Optional[float] = None
    ) -> None:
        """Create a new input consumer.

        Parameters
        ----------
        host : str
            The host to connect to.
        port : int
            The port to connect to.
        timeout : float, optional
            The timeout for the consumer, by default None (no timeout).
        """
        self.host = host
        self.port = port
        self.timeout = timeout
        self._running = False
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)

    def __enter__(self) -> "TCPConsumer":
        """Enter the context."""
        self.start()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Exit the context."""
        self.stop()

    def is_running(self) -> bool:
        """Check if the consumer is running.

        Returns
        -------
        bool
            True is the consumer is running, else False
        """
        return self._running

    def start(self) -> None:
        """Start the consumer."""
        if self._running:
            return
        self._running = True
        self.socket.connect((self.host, self.port))
        self.socket.sendall("CONSUMER\r\n".encode("utf-8"))

    def _get_response_no_timeout(self) -> Optional[str]:
        """Get the response."""
        data = self.socket.recv(1024)
        while not data.endswith(END_OF_MESSAGE):
            data += self.socket.recv(1024)
        if data.startswith(b"INPUT:"):
            response = data[len(b"INPUT:") :]
            if response.endswith(END_OF_MESSAGE):
                response = response[: -len(END_OF_MESSAGE)]
            return response.decode("utf-8")
        return None

    def _get_response_with_timeout(self, timeout: float) -> Optional[str]:
        """Get the response using a timeout."""
        start_time = time.monotonic()
        data = b""
        self.socket.settimeout(timeout)
        while time.monotonic() - start_time < timeout:
            try:
                data += self.socket.recv(1024)
            except TimeoutError:
                return None
            if data.endswith(END_OF_MESSAGE):
                break
        if not data:
            return None
        if data.startswith(b"INPUT:"):
            response = data[len(b"INPUT:") :]
            if response.endswith(END_OF_MESSAGE):
                response = response[: -len(END_OF_MESSAGE)]
            return response.decode("utf-8")
        return None

    def get_response(self) -> Optional[str]:
        """Get the response.

        Returns
        -------
        Optional[str]
            The response if available, None otherwise.
        """
        if not self._running:
            self.start()
        if self.timeout is None or self.timeout < 1:
            return self._get_response_no_timeout()
        return self._get_response_with_timeout(self.timeout)

    def send_prompt(self, prompt: str) -> None:
        """Send a prompt.

        Parameters
        ----------
        prompt : str
            The prompt to send.
        """
        if not self._running:
            self.start()
        message = f"REQUEST:{prompt}" + "\r\n"
        self.socket.sendall(message.encode("utf-8"))

    def stop(self) -> None:
        """Close the consumer."""
        try:
            self.socket.close()
        except OSError:
            pass
        del self.socket
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        self._running = False



"""Simple TCP server using twisted.

It listens for connections from an input provider and an input consumer,
and forwards messages between them.
"""

# pylint: disable=import-outside-toplevel,no-member,reimported,unused-import,redefined-outer-name,invalid-name  # noqa
import logging
import sys
from threading import Thread
from types import TracebackType
from typing import Dict, Optional, Type, cast

from twisted.internet.error import ReactorNotRestartable
from twisted.internet.interfaces import IReactorCore
from twisted.internet.protocol import Factory, Protocol, connectionDone
from twisted.internet.tcp import Port
from twisted.python.failure import Failure

LOGGER = logging.getLogger("tcp::server")
END_OF_MESSAGE = b"\r\n"


class ServerProtocol(Protocol):
    """Server protocol."""

    factory: "ServerFactory"

    def set_factory(self, factory: "ServerFactory") -> None:
        """Set the factory.

        Parameters
        ----------
        factory : ServerFactory
            The factory to set.
        """
        self.factory = factory

    def connectionLost(self, reason: Failure = connectionDone) -> None:
        """Handle connection lost event.

        Parameters
        ----------
        reason : Failure, optional
            The reason for the connection loss, by default connectionDone
        """
        if self.factory.clients["provider"] == self:
            self.factory.clients["provider"] = None
            LOGGER.info("Input provider disconnected.")
        elif self.factory.clients["consumer"] == self:
            self.factory.clients["consumer"] = None
            LOGGER.info("Input consumer disconnected.")
        super().connectionLost(reason)

    def message_received(self, message: str) -> None:
        """Handle a message received event.

        Parameters
        ----------
        message : str
            The message received.
        """
        if message.startswith("REQUEST:"):
            prompt = message[len("REQUEST:") :]
            if prompt.endswith("\r\n"):
                prompt = prompt[: -(len("\r\n"))]
            prompt = prompt.strip()
            LOGGER.debug("Received request: %s", prompt)
            if self.factory.clients["provider"]:
                msg = f"PROVIDE:{prompt}" + "\r\n"
                transport = self.factory.clients["provider"].transport
                transport.write(msg.encode("utf-8"))  # type: ignore
            else:
                LOGGER.error("No provider connected.")
        elif message.startswith("USE:"):
            response = message[len("USE:") :]
            if response.endswith("\r\n"):
                response = response[: -(len("\r\n"))]
            response = response.strip()
            LOGGER.debug("Received response: %s", response)
            if self.factory.clients["consumer"]:
                msg = f"INPUT:{response}" + "\r\n"
                transport = self.factory.clients["consumer"].transport
                transport.write(msg.encode("utf-8"))  # type: ignore
            else:
                LOGGER.error("No consumer connected.")

    def dataReceived(self, data: bytes) -> None:
        """Handle a data received event.

        Parameters
        ----------
        data : bytes
            The data received.
        """
        # we might get multiple messages in one chunk
        # i.e. CONSUMER\r\nREQUEST:prompt\r\n
        message = data.decode("utf-8")
        if message in ("PROVIDER\r\n", "PROVIDER\n", "PROVIDER"):
            LOGGER.debug("Input provider connected.")
            self.factory.clients["provider"] = self
            return
        if message.startswith("CONSUMER\r\n"):
            LOGGER.debug("Input consumer connected.")
            self.factory.clients["consumer"] = self
            rest = message[len("CONSUMER\r\n") :]
            if rest:
                self.message_received(rest)
            return
        self.message_received(message)


class ServerFactory(Factory):
    """Server factory."""

    protocol: "ServerProtocol"  # type: ignore
    clients: Dict[str, Optional["ServerProtocol"]]

    def __init__(self) -> None:
        """Initialize the factory."""
        super().__init__()
        self.clients = {
            "provider": None,
            "consumer": None,
        }

    def buildProtocol(self, addr: str) -> "ServerProtocol":
        """Build the protocol.

        Parameters
        ----------
        addr : str
            The address (ignored)

        Returns
        -------
        ServerProtocol
            The factory's protocol.
        """
        self.protocol = ServerProtocol()
        self.protocol.set_factory(self)
        return self.protocol


def get_reactor() -> IReactorCore:
    """Get the reactor from twisted.

    Returns
    -------
    IReactorCore
        The twisted's reactor
    """
    # dummy hack to allow restarting the reactor
    if "twisted.internet.reactor" in sys.modules:
        del sys.modules["twisted.internet.reactor"]
    import twisted.internet.error
    from twisted.internet import reactor  # noqa
    from twisted.internet import default

    try:
        default.install()
    # pylint: disable=line-too-long
    except (
        twisted.internet.error.ReactorAlreadyInstalledError
    ):  # pragma: no cover
        pass
    # cast it so mypy doesn't complain a lot
    reactor_cast = cast(IReactorCore, reactor)
    return reactor_cast


class TCPServerThread(Thread):
    """Threaded TCP server."""

    reactor: Optional[IReactorCore] = None  # noqa
    factory: Optional[Factory] = None  # noqa

    def __init__(
        self,
        interface: str,
        port: int,
        timeout: Optional[float] = None,
    ) -> None:
        """Create a new TCP server.

        Parameters
        ----------
        interface : str
            Interface to listen on. Defaults to '' (all interfaces)
        port : int
            Port to listen on.
        timeout : Optional[float]
            Timeout for the server.
        """
        super().__init__(
            name="TCPServerThread",
            daemon=True,
            target=self.run,
        )
        from twisted.internet.endpoints import TCP4ServerEndpoint

        self.timeout = timeout
        self.reactor = get_reactor()
        self._port = port
        endpoint = TCP4ServerEndpoint(  # type: ignore[no-untyped-call]
            self.reactor,
            port,
            interface=interface,
        )
        server_factory = ServerFactory()
        deferred = endpoint.listen(server_factory)  # type: ignore
        deferred.addCallback(callback=self.on_start)

    @property
    def port(self) -> int:
        """Get the port."""
        return self._port

    def on_start(self, port: Port) -> None:
        """On connect callback.

        Parameters
        ----------
        port : Port
            The port to connect to.
        """
        socket = port.getHost()  # type: ignore[no-untyped-call]
        LOGGER.debug(
            "listening on %s:%s",
            socket.host,
            socket.port,
        )
        self._port = socket.port
        self.factory = port.factory

    def run(self) -> None:
        """Start the server.

        Raises
        ------
        RuntimeError
            If reactor is not initialized
        """
        if self.reactor is None:  # pragma: no cover (just for the linter)
            raise RuntimeError("reactor is not running")
        if not self.reactor.running:
            try:
                self.reactor.run(installSignalHandlers=False)  # type: ignore
            except ReactorNotRestartable:  # pragma: no cover
                self.reactor = get_reactor()
                self.reactor.run(installSignalHandlers=False)  # type: ignore


class ServerWrapper:
    """Server Wrapper."""

    server: TCPServerThread
    timeout: float

    def __init__(
        self,
        interface: str,
        port: int,
        timeout: Optional[float] = None,
    ) -> None:
        """Create a new TCP server.

        Parameters
        ----------
        interface : str
            Interface to listen on. Defaults to '' (all interfaces)
        port : int
            Port to listen on.
        """
        self.timeout = timeout if timeout is not None else 0.2
        self.server = TCPServerThread(
            interface=interface, port=port, timeout=self.timeout
        )

    @property
    def port(self) -> int:
        """Get the port.

        Raises
        ------
        RuntimeError
            If the server is not running
        """
        if self.server.factory is None:
            raise RuntimeError("server is not running")
        return self.server.port

    def start(self) -> None:
        """Start the server.

        Raises
        ------
        RuntimeError
            If the server is not running
        """
        if self.server is None:
            raise RuntimeError("server is not running")
        self.server.start()

    def stop(self) -> None:
        """Stop the server."""
        # pylint: disable=line-too-long
        self.server.reactor.callFromThread(self.server.reactor.stop)  # type: ignore  # noqa
        self.server.join()


class TCPServer:
    """TCP Server."""

    _wrapper: Optional[ServerWrapper] = None

    def __init__(
        self,
        port: int,
        timeout: Optional[float] = None,
        interface: str = "",
    ) -> None:
        """Create a new server.

        Parameters
        ----------
        port : int
            Port to listen on.
        timeout : Optional[float]
            Timeout for the server.
        interface : str
            Interface to listen on. Defaults to '' (all interfaces)
        """
        self._port = port
        self._timeout = timeout
        self._interface = interface
        self._init_wrapper()
        self._running = False

    @property
    def port(self) -> int:
        """Get the port."""
        if self._wrapper is None:
            return self._port
        return self._wrapper.port

    def _init_wrapper(self) -> None:
        """Initialize the wrapper."""
        self._wrapper = ServerWrapper(
            port=self._port,
            timeout=self._timeout,
            interface=self._interface,
        )

    def start(self) -> None:
        """Start the server.

        Raises
        ------
        RuntimeError
            If the wrapper is not initialized
        """
        if self._running:
            return
        if not self._wrapper:
            self._init_wrapper()
        if not self._wrapper:  # pragma: no cover (just for the linter)
            raise RuntimeError("Server wrapper is not initialized")
        self._wrapper.start()
        self._port = self._wrapper.port
        self._running = True

    def stop(self) -> None:
        """Stop the server."""
        if not self._running:
            return
        if not self._wrapper:  # pragma: no cover (just for the linter)
            return
        self._wrapper.stop()
        self._running = False
        del self._wrapper
        self._wrapper = None

    def is_running(self) -> bool:
        """Check if the server is running.

        Returns
        -------
        bool
            True if the server is running, else False.
        """
        return self._running

    def __enter__(self) -> "TCPServer":
        """Enter the context manager."""
        self.start()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Exit the context manager."""
        self.stop()

    def restart(self) -> None:
        """Restart the server."""
        self.stop()
        self._init_wrapper()
        self.start()



"""Custom IOStream class to use with autogen.

It is meant to be used when we want to use custom
`print` and `input`. For example, when a websocket
is used to trigger a UI element that requires user input.
and sends back the user's input to the websocket. In the same
way, we can use it to forward what is meant to be printed.


We use:

- one tcp server to handle messaging between the clients
- one tcp client (provider) to set and forward the user's input
    that we got elsewhere (e.g. from a websocket connection)
- one tcp client (consumer) to ask and get the input from the provider
"""

import socket
from contextlib import closing
from typing import Any, Callable, Optional

from autogen.io import IOStream  # type: ignore[import-untyped]

from .stream import TCPConsumer, TCPProvider, TCPServer


class WaldiezIOStream(IOStream):
    """Custom IOStream class to handle the `print` and `input` functions."""

    def __init__(
        self,
        port: int = 0,
        input_timeout: float = 60,
        print_function: Optional[Callable[..., None]] = None,
        on_prompt_input: Optional[Callable[[str], None]] = None,
    ) -> None:
        """Initialize the IOStream.

        Parameters
        ----------
        port : int, optional
            The port to use, by default 0 (auto-assign).
        input_timeout : float, optional
            The input timeout, by default 60.
        print_function : Optional[Callable[..., None]], optional
            The print function to use, by default None.
        on_prompt_input : Optional[Callable[[str], None]], optional
            The function to call for getting an input, by default None.
        """
        self._print_function = print_function
        if port == 0:
            port = get_available_port()
        self._port = port
        self._input_timeout = input_timeout
        self._server = TCPServer(port)
        self._server.start()
        self._provider = TCPProvider("localhost", port, response=None)
        self._on_prompt_input = on_prompt_input

    @property
    def print_function(self) -> Optional[Callable[..., None]]:
        """Get the print function."""
        return self._print_function

    def open(self) -> None:
        """Start the server."""
        if not self._server.is_running():
            self._server.start()

    def close(self) -> None:
        """Stop the server and the provider."""
        # pylint: disable=broad-except
        if self._server.is_running():
            try:
                self._server.stop()
            except BaseException:  # pragma: no cover
                pass
        try:
            self._provider.stop()
        except BaseException:  # pragma: no cover
            pass

    def __del__(self) -> None:  # pragma: no cover
        """Delete the instance."""
        self.close()

    def forward_input(self, input_data: str) -> None:
        """Forward the user's input to the provider.

        When we have the input data
            e.g. from 'input(..)' or from a websocket connection,
            we can forward it to the provider (the tcp client)
            to make it available to the consumer (the other tcp client).
        Parameters
        ----------
        input_data : str
            The input data to forward.
        """
        if not self._provider.is_running():
            self._provider.start()
        self._provider.set_response(input_data)

    def print(
        self,
        *objects: Any,
        sep: str = " ",
        end: str = "\n",
        flush: bool = False,
    ) -> None:
        """Mock the print function.

        Parameters
        ----------
        objects : Any
            The objects to print.
        sep : str, optional
            The separator, by default " ".
        end : str, optional
            The end, by default a new line.
        flush : bool, optional
            Whether to flush, by default False.
        """
        print_function: Callable[..., None] = self.print_function or print
        print_function(*objects, sep=sep, end=end, flush=flush)

    def input(self, prompt: str = "", *, password: bool = False) -> str:
        """Mock the input function.

        Parameters
        ----------
        prompt : str, optional
            The prompt to show, by default "".
        password : bool, optional (not used)
            Whether to show the input as password, by default False.

        Returns
        -------
        str
            The user's input.
        """
        _prompt = prompt or "Your input:"
        if _prompt in (">", "> "):
            _prompt = "Your input:"
        if prompt:
            if self._on_prompt_input:
                self._on_prompt_input(_prompt)
            self.print(_prompt, end="")
        if not self._provider.is_running():
            self._provider.start()
        # wait for the provider to start
        self._provider.wait(timeout=self._input_timeout)
        if not self._provider.is_running():  # pragma: no cover
            self.print(
                "WARNING: Provider is not running. Was an input expected?"
            )
            return "\n"
        consumer = TCPConsumer(
            "localhost", self._port, timeout=self._input_timeout
        )
        consumer.start()
        # send the prompt and wait for the response
        consumer.send_prompt(_prompt)
        response = consumer.get_response()
        consumer.stop()
        self._provider.stop()
        # return the response or a line break (i.e. no input)
        return response or "\n"


def get_available_port() -> int:
    """Get an available port.

    Returns
    -------
    int
        Available port.
    """
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as soc:
        soc.bind(("", 0))
        soc.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return soc.getsockname()[1]



"""Command line interface to convert or run a waldiez file."""

import argparse
import json
import logging
import os
import sys
from pathlib import Path
from typing import Any, Dict, Optional

from autogen import ChatResult  # type: ignore[import-untyped]

from . import Waldiez, __version__
from .exporter import WaldiezExporter
from .runner import WaldiezRunner


def get_parser() -> argparse.ArgumentParser:
    """Get the argument parser for the Waldiez package.

    Returns
    -------
    argparse.ArgumentParser
        The argument parser.
    """
    parser = argparse.ArgumentParser(
        description="Run or export a Waldiez flow.",
        prog="waldiez",
    )
    parser.add_argument(
        "waldiez",
        type=str,
        help="Path to the Waldiez flow (*.waldiez) file.",
    )
    parser.add_argument(
        "-e",
        "--export",
        action="store_true",
        help=(
            "Export the Waldiez flow to a Python script or a jupyter notebook."
        ),
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help=(
            "Path to the output file. "
            "If exporting, the file extension determines the output format. "
            "If running, the output's directory will contain "
            "the generated flow (.py) and any additional generated files."
        ),
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help=("Override the output file if it already exists. "),
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"waldiez version: {__version__}",
    )
    return parser


def _log_result(result: ChatResult) -> None:
    """Log the result of the Waldiez flow."""
    logger = logging.getLogger("waldiez::cli")
    logger.info("Chat History:\n")
    logger.info(result.chat_history)
    logger.info("Summary:\n")
    logger.info(result.summary)
    logger.info("Cost:\n")
    logger.info(result.cost)


def _run(data: Dict[str, Any], output_path: Optional[str]) -> None:
    """Run the Waldiez flow."""
    waldiez = Waldiez.from_dict(data)
    runner = WaldiezRunner(waldiez)
    results = runner.run(stream=None, output_path=output_path)
    if isinstance(results, list):
        for result in results:
            _log_result(result)
            sep = "-" * 80
            print(f"\n{sep}\n")
    else:
        _log_result(results)


def main() -> None:
    """Parse the command line arguments and run the Waldiez flow."""
    parser = get_parser()
    args = parser.parse_args()
    logger = _get_logger()
    waldiez_file: str = args.waldiez
    if not os.path.exists(waldiez_file):
        logger.error("File not found: %s", waldiez_file)
        sys.exit(1)
    if not waldiez_file.endswith((".json", ".waldiez")):
        logger.error("Only .json or .waldiez files are supported.")
        sys.exit(1)
    with open(waldiez_file, "r", encoding="utf-8") as file:
        try:
            data = json.load(file)
        except json.decoder.JSONDecodeError:
            logger.error("Invalid .waldiez file: %s. Not a valid json?", file)
            return
    if args.export is True:
        if args.output is None:
            logger.error("Please provide an output file.")
            sys.exit(1)
        if not args.output.endswith((".py", ".ipynb", ".json", ".waldiez")):
            logger.error(
                "Only Python scripts, Jupyter notebooks "
                "and JSON/Waldiez files are supported."
            )
            sys.exit(1)
        output_file = Path(args.output).resolve()
        waldiez = Waldiez.from_dict(data)
        exporter = WaldiezExporter(waldiez)
        exporter.export(output_file, force=args.force)
        generated = str(output_file).replace(os.getcwd(), ".")
        logger.info("Generated: %s", generated)
    else:
        _run(data, args.output)


def _get_logger(level: int = logging.INFO) -> logging.Logger:
    """Get the logger for the Waldiez package.

    Parameters
    ----------
    level : int or str, optional
        The logging level. Default is logging.INFO.

    Returns
    -------
    logging.Logger
        The logger.
    """
    # check if we already have setup a config

    if not logging.getLogger().handlers:
        logging.basicConfig(
            level=level,
            format="%(levelname)s %(message)s",
            stream=sys.stderr,
            force=True,
        )
    logger = logging.getLogger("waldiez::cli")
    current_level = logger.getEffectiveLevel()
    if current_level != level:
        logger.setLevel(level)
    return logger


if __name__ == "__main__":
    main()



"""Waldiez package."""

from ._version import __version__
from .exporter import WaldiezExporter
from .io_stream import WaldiezIOStream
from .models import Waldiez
from .runner import WaldiezRunner

__all__ = [
    "Waldiez",
    "WaldiezExporter",
    "WaldiezIOStream",
    "WaldiezRunner",
    "__version__",
]
